# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved

from __future__ import annotations

import time
from enum import Enum, auto
from typing import Callable, Type

import numpy as np

from PySide6.QtGui import QFont
from PySide6.QtWidgets import QWidget

import pyqtgraph as pg

import acconeer.exptool as et
from acconeer.exptool import a121
from acconeer.exptool.a121.algo._plugins import (
    ProcessorBackendPluginBase,
    ProcessorPlotPluginBase,
    ProcessorPluginPreset,
    ProcessorViewPluginBase,
)
from acconeer.exptool.a121.algo.breathing import (
    Processor,
    ProcessorConfig,
    ProcessorResult,
    get_sensor_config,
)
from acconeer.exptool.app.new import (
    AppModel,
    Message,
    PidgetFactoryMapping,
    PluginFamily,
    PluginGeneration,
    PluginPresetBase,
    PluginSpecBase,
    pidgets,
)


class PluginPresetId(Enum):
    DEFAULT = auto()


class BackendPlugin(ProcessorBackendPluginBase[ProcessorConfig, ProcessorResult]):

    PLUGIN_PRESETS = {
        PluginPresetId.DEFAULT.value: lambda: ProcessorPluginPreset(
            session_config=a121.SessionConfig(get_sensor_config()),
            processor_config=BackendPlugin.get_processor_config_cls()(),
        ),
    }

    @classmethod
    def get_processor_cls(cls) -> Type[Processor]:
        return Processor

    @classmethod
    def get_processor_config_cls(cls) -> Type[ProcessorConfig]:
        return ProcessorConfig

    @classmethod
    def get_default_sensor_config(cls) -> a121.SensorConfig:
        return get_sensor_config()


class ViewPlugin(ProcessorViewPluginBase[ProcessorConfig]):
    @classmethod
    def get_pidget_mapping(cls) -> PidgetFactoryMapping:
        # Note: Incomplete mapping
        return {
            "time_series_length": pidgets.IntParameterWidgetFactory(
                name_label_text="Time series length",
                limits=(0, None),
            ),
            "lp_coeff": pidgets.FloatSliderParameterWidgetFactory(
                name_label_text="Time filtering coefficient",
                limits=(0, 1),
                decimals=2,
            ),
            "min_freq": pidgets.FloatSliderParameterWidgetFactory(
                name_label_text="Lower frequency of bandpass filter",
                limits=(0, 1.0),
                decimals=1,
            ),
            "max_freq": pidgets.FloatSliderParameterWidgetFactory(
                name_label_text="Upper frequency of bandpass filter",
                limits=(2.0, 10.0),
                decimals=1,
            ),
        }

    @classmethod
    def get_processor_config_cls(cls) -> Type[ProcessorConfig]:
        return ProcessorConfig


class PlotPlugin(ProcessorPlotPluginBase[ProcessorResult]):

    _TIME_SERIES_Y_SCALE_MARGIN_M = 0.0025
    _DATA_INIT_DURATION = 5.0

    def __init__(self, *, plot_layout: pg.GraphicsLayout, app_model: AppModel) -> None:
        super().__init__(plot_layout=plot_layout, app_model=app_model)

    def setup(self, metadata: a121.Metadata, sensor_config: a121.SensorConfig) -> None:

        pens = [et.utils.pg_pen_cycler(i) for i in range(5)]
        brush = et.utils.pg_brush_cycler(0)
        symbol_kw = dict(symbol="o", symbolSize=1, symbolBrush=brush, symbolPen="k")
        feat_kws = [dict(pen=pen, **symbol_kw) for pen in pens]

        font = QFont()
        font.setPixelSize(16.0)

        # time series plot
        self.time_series_plot = self.plot_layout.addPlot(row=0, col=0)
        self.time_series_plot.setMenuEnabled(False)
        self.time_series_plot.showGrid(x=True, y=True)
        self.time_series_plot.addLegend()
        self.time_series_plot.setLabel("left", "Displacement (m)")
        self.time_series_plot.setLabel("bottom", "Time")
        self.time_series_plot.addItem(pg.PlotDataItem())
        self.time_series_curve = []
        self.time_series_curve.append(self.time_series_plot.plot(**feat_kws[0]))
        self.time_series_curve.append(self.time_series_plot.plot(**feat_kws[1]))

        self.time_series_smooth_limits = et.utils.SmoothLimits()

        self.time_series_text_item = pg.TextItem(
            fill=pg.mkColor(0xFF, 0x7F, 0x0E, 200),
            anchor=(0.5, 0),
            color=pg.mkColor(0xFF, 0xFF, 0xFF, 200),
        )
        self.time_series_text_item.setFont(font)
        self.time_series_text_item.show()
        self.time_series_plot.addItem(self.time_series_text_item)

        # fft plot
        self.fft_plot = self.plot_layout.addPlot(row=1, col=0)
        self.fft_plot.setMenuEnabled(False)
        self.fft_plot.showGrid(x=True, y=True)
        self.fft_plot.addLegend()
        self.fft_plot.setLabel("left", "Power")
        self.fft_plot.setLabel("bottom", "Frequency (Hz)")
        self.fft_plot.addItem(pg.PlotDataItem())
        self.fft_curve = self.fft_plot.plot(**feat_kws[0])
        self.fft_vert_line = pg.InfiniteLine(pen=pens[1])
        self.fft_plot.addItem(self.fft_vert_line)

        self.fft_smooth_max = et.utils.SmoothMax(tau_grow=0.5, tau_decay=2.0)

        self.fft_text_item = pg.TextItem(
            fill=pg.mkColor(0xFF, 0x7F, 0x0E, 200),
            anchor=(0.5, 0),
            color=pg.mkColor(0xFF, 0xFF, 0xFF, 200),
        )
        self.fft_text_item.setFont(font)
        self.fft_text_item.show()
        self.fft_plot.addItem(self.fft_text_item)

        self.start_time = time.time()

    def update(self, processor_result: ProcessorResult) -> None:

        time_series = processor_result.time_series
        distance_to_analyze_m = processor_result.distance_to_analyze_m
        z_abs_lp = processor_result.lp_psd
        freqs = processor_result.freqs
        breathing_rate = processor_result.breathing_rate
        fft_peak_location = processor_result.fft_peak_location

        # time series plot
        self.time_series_curve[0].setData(time_series)
        lim = self.time_series_smooth_limits.update(time_series)

        self.time_series_plot.setYRange(
            lim[0] - self._TIME_SERIES_Y_SCALE_MARGIN_M,
            lim[1] + self._TIME_SERIES_Y_SCALE_MARGIN_M,
        )
        text_y_pos = self.time_series_plot.getAxis("left").range[1] * 0.95
        text_x_pos = self.time_series_plot.getAxis("bottom").range[1] / 2.0
        self.time_series_text_item.setPos(text_x_pos, text_y_pos)
        self.time_series_text_item.setHtml(
            "Distance being visualized: " + "{:.2f}".format(distance_to_analyze_m) + " (m)"
        )

        # fft plot
        if (time.time() - self.start_time) < self._DATA_INIT_DURATION:
            self.fft_curve.setData(freqs, np.zeros_like(freqs))
            self.fft_plot.setYRange(0.0, 1.0)

            text_y_pos = self.fft_plot.getAxis("left").range[1] * 0.95
            text_x_pos = self.fft_plot.getAxis("bottom").range[1] / 2.0
            self.fft_text_item.setHtml("Acquiring data.")
        elif freqs is not None:
            assert z_abs_lp is not None
            assert breathing_rate is not None
            assert fft_peak_location is not None

            self.fft_curve.setData(freqs, z_abs_lp)
            self.fft_vert_line.setValue(fft_peak_location)
            lim = self.fft_smooth_max.update(z_abs_lp)
            self.fft_plot.setYRange(0.0, lim * 1.2)

            text_y_pos = self.fft_plot.getAxis("left").range[1] * 0.95
            text_x_pos = self.fft_plot.getAxis("bottom").range[1] / 2.0

            self.fft_text_item.setHtml(
                "Breathing rate: " + "{:.1f}".format(breathing_rate) + " per minute"
            )
        self.fft_text_item.setPos(text_x_pos, text_y_pos)

    @staticmethod
    def _format_string(string_to_display: str) -> str:
        return (
            '<div style="text-align: center">'
            '<span style="color: #FFFFFF;font-size:15pt;">'
            "{}</span></div>".format(string_to_display)
        )


class PluginSpec(PluginSpecBase):
    def create_backend_plugin(
        self, callback: Callable[[Message], None], key: str
    ) -> BackendPlugin:
        return BackendPlugin(callback=callback, generation=self.generation, key=key)

    def create_view_plugin(self, app_model: AppModel, view_widget: QWidget) -> ViewPlugin:
        return ViewPlugin(app_model=app_model, view_widget=view_widget)

    def create_plot_plugin(
        self, app_model: AppModel, plot_layout: pg.GraphicsLayout
    ) -> PlotPlugin:
        return PlotPlugin(app_model=app_model, plot_layout=plot_layout)


BREATHING_PLUGIN = PluginSpec(
    generation=PluginGeneration.A121,
    key="breathing",
    title="Breathing",
    description="Measure breathing rate.",
    family=PluginFamily.EXAMPLE_APP,
    presets=[
        PluginPresetBase(name="Default", preset_id=PluginPresetId.DEFAULT),
    ],
    default_preset_id=PluginPresetId.DEFAULT,
)



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved

from __future__ import annotations

from typing import Optional

import attrs
import numpy as np
import numpy.typing as npt
from scipy import signal

from acconeer.exptool import a121
from acconeer.exptool.a121.algo import (
    APPROX_BASE_STEP_LENGTH_M,
    AlgoProcessorConfigBase,
    ProcessorBase,
)


@attrs.mutable(kw_only=True)
class ProcessorConfig(AlgoProcessorConfigBase):
    time_series_length: int = attrs.field(default=1024)
    """Length of time series."""

    lp_coeff: float = attrs.field(default=0.75)
    """Specify filter coefficient for exponential filter of psd over time."""

    min_freq: float = attrs.field(default=0.2)
    """Lower limit of bandpass filter."""

    max_freq: float = attrs.field(default=3.0)
    """Upper limit of bandpass filter."""

    def _collect_validation_results(
        self, config: a121.SessionConfig
    ) -> list[a121.ValidationResult]:
        validation_results: list[a121.ValidationResult] = []

        if config.sensor_config.frame_rate is None:
            validation_results.append(
                a121.ValidationError(
                    config.sensor_config,
                    "frame_rate",
                    "Must be set",
                )
            )

        if config.sensor_config.sweeps_per_frame != 1:
            validation_results.append(
                a121.ValidationError(
                    config.sensor_config,
                    "sweeps_per_frame",
                    "Must be equal to 1",
                )
            )

        return validation_results


@attrs.frozen(kw_only=True)
class ProcessorContext:
    ...


@attrs.frozen(kw_only=True)
class ProcessorResult:
    time_series: npt.NDArray[np.float_]
    distance_to_analyze_m: float
    freqs: npt.NDArray[np.float_]
    lp_psd: Optional[npt.NDArray[np.float_]] = attrs.field(default=None)
    breathing_rate: Optional[float] = attrs.field(default=None)
    fft_peak_location: Optional[float] = attrs.field(default=None)


class Processor(ProcessorBase[ProcessorConfig, ProcessorResult]):

    _IIR_INIT_LENGTH: int = 200
    _OVER_SAMPLING_FACTOR: int = 5
    _PHASE_TO_MM: float = APPROX_BASE_STEP_LENGTH_M / (np.pi * 2.0)
    _SECONDS_PER_MINUTE: float = 60.0
    _SECONDS_PER_FFT_WINDOW: float = 10.0
    _AMPL_LP_COEFF: float = 0.99

    def __init__(
        self,
        *,
        sensor_config: a121.SensorConfig,
        metadata: a121.Metadata,
        processor_config: ProcessorConfig,
        subsweep_indexes: Optional[list[int]] = None,
        context: Optional[ProcessorContext] = None,
    ) -> None:
        processor_config.validate(sensor_config)

        assert sensor_config.frame_rate is not None
        self.frame_rate = sensor_config.frame_rate
        self.start_point = sensor_config.start_point
        self.step_length = sensor_config.step_length
        self.spf = sensor_config.sweeps_per_frame

        self.time_series_len = processor_config.time_series_length
        self.max_freq = processor_config.max_freq
        self.lp_coeff = processor_config.lp_coeff

        self.time_series = np.zeros(
            shape=(
                processor_config.time_series_length + self._IIR_INIT_LENGTH,
                sensor_config.num_points,
            )
        )

        self.has_init = False

        self.ampl_lp_filt = np.zeros(shape=sensor_config.num_points)

        self.lp_psd = None
        self.window_length = self._SECONDS_PER_FFT_WINDOW * self.frame_rate
        (self.B, self.A) = signal.butter(
            N=2,
            Wn=[processor_config.min_freq, processor_config.max_freq],
            fs=sensor_config.frame_rate,
            btype="bandpass",
        )

    def process(self, result: a121.Result) -> ProcessorResult:
        # Add new data to fifo buffer, unwrap and filter.
        self.time_series = np.roll(self.time_series, -self.spf, axis=0)
        self.time_series[-self.spf :] = np.angle(result.frame)
        time_series_unwrapped = np.unwrap(self.time_series, axis=0) * self._PHASE_TO_MM
        time_series_filt = signal.lfilter(self.B, self.A, time_series_unwrapped, axis=0)[
            self._IIR_INIT_LENGTH :
        ]

        # Filter amplitudes and determine at what distance to plot time series.
        self.ampl_lp_filt = self._AMPL_LP_COEFF * self.ampl_lp_filt + (
            1 - self._AMPL_LP_COEFF
        ) * np.abs(result.frame)
        distance_to_plot = np.argmax(self.ampl_lp_filt)
        distance_to_plot_m = (
            float(self.start_point + distance_to_plot * self.step_length)
            * APPROX_BASE_STEP_LENGTH_M
        )

        # Calculate psd.
        freqs, psd = signal.welch(
            x=time_series_filt,
            fs=self.frame_rate,
            window="hann",
            nperseg=self.window_length,
            nfft=self.time_series_len * self._OVER_SAMPLING_FACTOR,
            axis=0,
            average="mean",
        )

        # Weigh psd with filtered amplitude and psd over distances.
        psd = np.mean(psd * self.ampl_lp_filt, axis=1)

        if not self.has_init:
            self.lp_psd = psd
            self.ampl_lp_filt = np.abs(result.frame)
            self.has_init = True

        assert self.lp_psd is not None
        self.lp_psd = self.lp_psd * self.lp_coeff + psd * (1 - self.lp_coeff)

        idx_to_analyze = np.where(freqs < self.max_freq)[0]
        freqs_to_analyze = freqs[idx_to_analyze]
        psd_to_analyze = self.lp_psd[idx_to_analyze]
        fft_peak_loc_idx = np.argmax(psd_to_analyze)

        # Make sure max value is not first or last element in psd.
        if (fft_peak_loc_idx != 0) & (fft_peak_loc_idx != (freqs_to_analyze.shape[0] - 1)):
            fft_interpolated_freq = self.interpolate_peaks(
                psd_to_analyze[fft_peak_loc_idx - 1 : fft_peak_loc_idx + 2],
                freqs_to_analyze[fft_peak_loc_idx - 1 : fft_peak_loc_idx + 2],
            )
            breathing_rate = fft_interpolated_freq * self._SECONDS_PER_MINUTE

            return ProcessorResult(
                time_series=time_series_filt[:, distance_to_plot],
                distance_to_analyze_m=distance_to_plot_m,
                lp_psd=psd_to_analyze,
                freqs=freqs_to_analyze,
                breathing_rate=breathing_rate,
                fft_peak_location=fft_interpolated_freq,
            )

        else:
            return ProcessorResult(
                time_series=time_series_filt[:, distance_to_plot],
                distance_to_analyze_m=distance_to_plot_m,
                freqs=freqs_to_analyze,
            )

    @staticmethod
    def interpolate_peaks(ampls: npt.NDArray[np.float_], freqs: npt.NDArray[np.float_]) -> float:
        x = freqs
        y = ampls
        a = (x[0] * (y[2] - y[1]) + x[1] * (y[0] - y[2]) + x[2] * (y[1] - y[0])) / (
            (x[0] - x[1]) * (x[0] - x[2]) * (x[1] - x[2])
        )
        b = (y[1] - y[0]) / (x[1] - x[0]) - a * (x[0] + x[1])
        return float(-b / (2 * a))

    def update_config(self, config: ProcessorConfig) -> None:
        ...


def get_sensor_config() -> a121.SensorConfig:
    return a121.SensorConfig(
        profile=a121.Profile.PROFILE_3,
        hwaas=16,
        num_points=10,
        step_length=24,
        start_point=150,
        receiver_gain=12,
        sweeps_per_frame=1,
        frame_rate=50,
    )



# Copyright (c) Acconeer AB, 2023
# All rights reserved
from __future__ import annotations

from typing import Optional

import attrs
import numpy as np
import numpy.typing as npt

from acconeer.exptool import a121
from acconeer.exptool.a121.algo import AlgoProcessorConfigBase
from acconeer.exptool.a121.algo.presence import DetectorConfig, DetectorResult


@attrs.frozen(kw_only=True)
class ProcessorResult:
    """Processor result"""

    zone_limits: npt.NDArray[np.float_] = attrs.field()
    max_presence_zone: Optional[int] = attrs.field()
    total_zone_detections: npt.NDArray[np.int_] = attrs.field()
    inter_zone_detections: npt.NDArray[np.int_] = attrs.field()
    max_inter_zone: Optional[int] = attrs.field()
    intra_zone_detections: npt.NDArray[np.int_] = attrs.field()
    max_intra_zone: Optional[int] = attrs.field()


@attrs.mutable(kw_only=True)
class ProcessorConfig(AlgoProcessorConfigBase):
    num_zones: int = attrs.field(default=3)

    def _collect_validation_results(
        self, config: a121.SessionConfig
    ) -> list[a121.ValidationResult]:

        return []


class Processor:
    """Smart presence processor
    :param processor_config: Processor configuration.
    """

    max_inter_zone: Optional[int]
    max_intra_zone: Optional[int]

    def __init__(
        self,
        processor_config: ProcessorConfig,
        detector_config: DetectorConfig,
        session_config: a121.SessionConfig,
    ):
        num_points = session_config.sensor_config.num_points
        self.num_zones = np.minimum(processor_config.num_zones, num_points)
        self.distances = np.linspace(detector_config.start_m, detector_config.end_m, num_points)
        self.zone_limits = self.create_zones(self.distances, self.num_zones)

        self.inter_enable = detector_config.inter_enable
        self.inter_threshold = detector_config.inter_detection_threshold
        self.inter_zones = np.zeros(self.num_zones, dtype=int)
        self.max_inter_zone = None

        self.intra_enable = detector_config.intra_enable
        self.intra_threshold = detector_config.intra_detection_threshold
        self.intra_zones = np.zeros(self.num_zones, dtype=int)
        self.max_intra_zone = None

    @staticmethod
    def create_zones(distances: npt.NDArray[np.float_], num_zones: int) -> npt.NDArray[np.float_]:
        """Create zones limits based on chosen range and number of zones."""

        if num_zones < distances.size:
            zone_limits = np.linspace(distances[0], distances[-1], num_zones + 1)
            # Only return upper limits
            return zone_limits[1:]  # type: ignore[no-any-return]

        else:
            return distances

    def get_zone_detections(
        self, depthwise_scores: npt.NDArray[np.float_], threshold: float
    ) -> npt.NDArray[np.int_]:
        """Get presence detection result for all zones."""

        zones = np.zeros(self.num_zones, dtype=int)
        zone_detected = False
        limit_idx = 0
        for i, score in enumerate(depthwise_scores):
            if self.distances[i] > self.zone_limits[limit_idx]:
                limit_idx += 1
                zone_detected = False
            if not zone_detected:
                zone_detected = score > threshold
                zones[limit_idx] = 1 if zone_detected else 0

        return zones

    def get_max_presence_zone(self, depthwise_scores: npt.NDArray[np.float_]) -> int:
        """Get the zone with maximum presence score."""

        max_idx = np.argmax(depthwise_scores)
        max_distance = self.distances[max_idx]
        max_zone = 0
        while max_zone < self.num_zones - 1 and max_distance > self.zone_limits[max_zone]:
            max_zone += 1

        return max_zone

    def process(self, result: DetectorResult) -> ProcessorResult:
        max_presence_zone: Optional[int]

        if result.presence_detected:
            # Update zone detections for each detection type.
            if self.inter_enable:
                if result.inter_presence_score > self.inter_threshold:
                    self.inter_zones = self.get_zone_detections(
                        result.inter_depthwise_scores, self.inter_threshold
                    )
                    if not np.any(self.inter_zones):
                        self.inter_zones[self.max_inter_zone] = 1
                    else:
                        self.max_inter_zone = self.get_max_presence_zone(
                            result.inter_depthwise_scores
                        )

                else:
                    self.inter_zones = np.zeros(self.num_zones, dtype=int)
                    self.max_inter_zone = None

            if self.intra_enable:
                if result.intra_presence_score > self.intra_threshold:
                    self.intra_zones = self.get_zone_detections(
                        result.intra_depthwise_scores, self.intra_threshold
                    )
                    if not np.any(self.intra_zones):
                        self.intra_zones[self.max_intra_zone] = 1
                    else:
                        self.max_intra_zone = self.get_max_presence_zone(
                            result.intra_depthwise_scores
                        )

                else:
                    self.intra_zones = np.zeros(self.num_zones, dtype=int)
                    self.max_intra_zone = None

            # max intra zone is prioritized due to faster reaction time.
            if self.max_intra_zone is not None:
                max_presence_zone = self.max_intra_zone
            else:
                max_presence_zone = self.max_inter_zone

            total_zone_detections = np.maximum(self.inter_zones, self.intra_zones)

        else:
            max_presence_zone = None
            self.inter_zones = np.zeros(self.num_zones, dtype=int)
            self.intra_zones = np.zeros(self.num_zones, dtype=int)
            total_zone_detections = np.zeros(self.num_zones, dtype=int)
            self.max_inter_zone = None
            self.max_intra_zone = None

        return ProcessorResult(
            zone_limits=self.zone_limits,
            max_presence_zone=max_presence_zone,
            total_zone_detections=total_zone_detections,
            inter_zone_detections=self.inter_zones,
            max_inter_zone=self.max_inter_zone,
            intra_zone_detections=self.intra_zones,
            max_intra_zone=self.max_intra_zone,
        )



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved

from __future__ import annotations

import time

import numpy as np

# Added here to force pyqtgraph to choose PySide
import PySide6  # noqa: F401
from PySide6.QtGui import QFont

import pyqtgraph as pg

import acconeer.exptool as et
from acconeer.exptool import a121
from acconeer.exptool.a121.algo.breathing import Processor, ProcessorConfig, get_sensor_config


sensor_config = get_sensor_config()


def main():
    args = a121.ExampleArgumentParser().parse_args()
    et.utils.config_logging(args)

    client = a121.Client(**a121.get_client_args(args))
    client.connect()
    metadata = client.setup_session(sensor_config)

    processor = Processor(
        sensor_config=sensor_config, processor_config=ProcessorConfig(), metadata=metadata
    )

    pg_updater = PGUpdater(sensor_config, metadata)
    pg_process = et.PGProcess(pg_updater)
    pg_process.start()

    client.start_session()
    interrupt_handler = et.utils.ExampleInterruptHandler()
    print("Press Ctrl-C to end session")

    while not interrupt_handler.got_signal:
        result = client.get_next()

        processor_result = processor.process(result)

        try:
            pg_process.put_data(processor_result)
        except et.PGProccessDiedException:
            break

    print("Disconnecting...")
    client.disconnect()


class PGUpdater:

    _TIME_SERIES_Y_SCALE_MARGIN_M = 0.0025
    _DATA_INIT_DURATION = 5.0

    def __init__(self, sensor_config: a121.SensorConfig, metadata: a121.Metadata):
        ...

    def setup(self, win):

        pens = [et.utils.pg_pen_cycler(i) for i in range(5)]
        brush = et.utils.pg_brush_cycler(0)
        symbol_kw = dict(symbol="o", symbolSize=1, symbolBrush=brush, symbolPen="k")
        feat_kws = [dict(pen=pen, **symbol_kw) for pen in pens]

        font = QFont()
        font.setPixelSize(16.0)

        # time series plot
        self.time_series_plot = win.addPlot(row=0, col=0)
        self.time_series_plot.setMenuEnabled(False)
        self.time_series_plot.showGrid(x=True, y=True)
        self.time_series_plot.addLegend()
        self.time_series_plot.setLabel("left", "Displacement (m)")
        self.time_series_plot.setLabel("bottom", "Time")
        self.time_series_plot.addItem(pg.PlotDataItem())
        self.time_series_curve = []
        self.time_series_curve.append(self.time_series_plot.plot(**feat_kws[0]))
        self.time_series_curve.append(self.time_series_plot.plot(**feat_kws[1]))

        self.time_series_smooth_limits = et.utils.SmoothLimits()

        self.time_series_text_item = pg.TextItem(
            fill=pg.mkColor(0xFF, 0x7F, 0x0E, 200),
            anchor=(0.5, 0),
            color=pg.mkColor(0xFF, 0xFF, 0xFF, 200),
        )
        self.time_series_text_item.setFont(font)
        self.time_series_text_item.show()
        self.time_series_plot.addItem(self.time_series_text_item)

        # fft plot
        self.fft_plot = win.addPlot(row=1, col=0)
        self.fft_plot.setMenuEnabled(False)
        self.fft_plot.showGrid(x=True, y=True)
        self.fft_plot.addLegend()
        self.fft_plot.setLabel("left", "Power")
        self.fft_plot.setLabel("bottom", "Frequency (Hz)")
        self.fft_plot.addItem(pg.PlotDataItem())
        self.fft_curve = self.fft_plot.plot(**feat_kws[0])
        self.fft_vert_line = pg.InfiniteLine(pen=pens[1])
        self.fft_plot.addItem(self.fft_vert_line)

        self.fft_smooth_max = et.utils.SmoothMax(tau_grow=0.5, tau_decay=2.0)

        self.fft_text_item = pg.TextItem(
            fill=pg.mkColor(0xFF, 0x7F, 0x0E, 200),
            anchor=(0.5, 0),
            color=pg.mkColor(0xFF, 0xFF, 0xFF, 200),
        )
        self.fft_text_item.setFont(font)
        self.fft_text_item.show()
        self.fft_plot.addItem(self.fft_text_item)

        self.start_time = time.time()

    def update(self, processor_result):
        time_series = processor_result.time_series
        distance_to_analyze_m = processor_result.distance_to_analyze_m
        z_abs_lp = processor_result.lp_psd
        freqs = processor_result.freqs
        breathing_rate = processor_result.breathing_rate
        fft_peak_location = processor_result.fft_peak_location

        # time series plot
        self.time_series_curve[0].setData(time_series)
        lim = self.time_series_smooth_limits.update(time_series)

        self.time_series_plot.setYRange(
            lim[0] - self._TIME_SERIES_Y_SCALE_MARGIN_M,
            lim[1] + self._TIME_SERIES_Y_SCALE_MARGIN_M,
        )
        text_y_pos = self.time_series_plot.getAxis("left").range[1] * 0.95
        text_x_pos = self.time_series_plot.getAxis("bottom").range[1] / 2.0
        self.time_series_text_item.setPos(text_x_pos, text_y_pos)
        self.time_series_text_item.setHtml(
            "Distance being visualized: " + "{:.2f}".format(distance_to_analyze_m) + " (m)"
        )

        # fft plot
        if (time.time() - self.start_time) < self._DATA_INIT_DURATION:
            self.fft_curve.setData(freqs, np.zeros_like(freqs))
            self.fft_plot.setYRange(0.0, 1.0)

            text_y_pos = self.fft_plot.getAxis("left").range[1] * 0.95
            text_x_pos = self.fft_plot.getAxis("bottom").range[1] / 2.0
            self.fft_text_item.setHtml("Acquiring data.")
        elif freqs is not None:
            assert z_abs_lp is not None
            assert breathing_rate is not None
            assert fft_peak_location is not None

            self.fft_curve.setData(freqs, z_abs_lp)
            self.fft_vert_line.setValue(fft_peak_location)
            lim = self.fft_smooth_max.update(z_abs_lp)
            self.fft_plot.setYRange(0.0, lim * 1.2)

            text_y_pos = self.fft_plot.getAxis("left").range[1] * 0.95
            text_x_pos = self.fft_plot.getAxis("bottom").range[1] / 2.0

            self.fft_text_item.setHtml(
                "Breathing rate: " + "{:.1f}".format(breathing_rate) + " per minute"
            )
        self.fft_text_item.setPos(text_x_pos, text_y_pos)

    @staticmethod
    def _format_string(string_to_display: str) -> str:
        return (
            '<div style="text-align: center">'
            '<span style="color: #FFFFFF;font-size:15pt;">'
            "{}</span></div>".format(string_to_display)
        )


if __name__ == "__main__":
    main()
